from typing import List
from dataclasses import dataclass

@dataclass(eq=True)
class Constraint:
    pass

@dataclass(eq=True)
class ForeignKeyConstraint(Constraint):
    references: str
    on_delete: str | None = None

@dataclass(eq=True)
class UniqueConstraint(Constraint):
    pass

@dataclass
class Column:
    name: str
    data_type: str
    primary_key: bool
    auto_increment: bool
    default: str | None
    constraints: List[ForeignKeyConstraint | UniqueConstraint]
    
    nullable: bool = False
    
    
@dataclass(eq=True)
class Index:
    name: str
    type: str
    columns: List[str]

@dataclass
class Table:
    name: str
    columns: List[Column]
    indexes: List[Index]

@dataclass
class Enum:
    name: str
    values: List[str]


import os
from dotenv import dotenv_values

def find_env_file():
    current_dir = os.getcwd()
    
    while True:
        # Check if requirements.txt exists in the current directory
        if os.path.exists(os.path.join(current_dir, 'requirements.txt')):
            # Look for .env file in this directory
            env_path = os.path.join(current_dir, '.env')
            if os.path.exists(env_path):
                return env_path
            else:
                raise FileNotFoundError("No .env file found in the same directory as requirements.txt")
        
        # Move up one directory
        current_dir = os.path.dirname(current_dir)
        
        # If we reach the root directory and haven't found requirements.txt, stop
        if current_dir == os.path.abspath(os.sep):
            raise FileNotFoundError("Could not find requirements.txt in any parent directory")

env_file_path = find_env_file()
config = dotenv_values(env_file_path)

def get_database_url():
    return config["DATABASE_URL"]


from dataclasses import dataclass
from typing import List, Any

@dataclass
class ModelColumn:
    table: str
    name: str

    def in_(self, values: List[Any]):
        from .conditions import Condition
        return Condition(column=self, condition="in", value=values)

    def __lt__(self, other):
        from .conditions import Condition
        return Condition(column=self, condition="<", value=other)

    def __gt__(self, other):
        from .conditions import Condition
        return Condition(column=self, condition=">", value=other)

    def __eq__(self, other):
        from .conditions import Condition
        return Condition(column=self, condition="=", value=other)


import importlib.metadata
import os
import sys
import importlib
from ...model import Model

# Add the FastAPI project's root directory to sys.path
project_root = os.getcwd()
if project_root not in sys.path:
    sys.path.insert(0, project_root)

def get_models():
    model_files = os.listdir("database/models")

    models = set()
    for model_file in model_files:
        if model_file.endswith(".py") == False:
            continue

        module = importlib.import_module(
            f"database.models.{model_file.replace(".py", "")}"
        )
        for attribute_name in dir(module):
            export = getattr(module, attribute_name)
            if (
                isinstance(export, type)
                and export != Model
                and issubclass(export, Model)
            ):
                models.add(export)

    return models



from typing import List, Optional, Tuple
from dataclasses import dataclass
from .schema import Table, Column, Constraint, Enum
from .action import TableAction, EnumAction

def get_schema_diff_actions(model_schema: Tuple[List[Table], List[Enum]], database_schema: Tuple[List[Table], List[Enum]]):
    actions = []

    model_tables, model_enums = model_schema
    database_tables, database_enums = database_schema

    for model_enum in model_enums:
        if model_enum.name not in [enum.name for enum in database_enums]:
            actions.append(EnumAction(type="CREATE_ENUM", enum=model_enum))

    for database_enum in database_enums:
        if database_enum.name not in [enum.name for enum in model_enums]:
            actions.append(EnumAction(type="DROP_ENUM", enum=database_enum))

    for enum in model_enums:
        database_enum = next((database_enum for database_enum in database_enums if database_enum.name == enum.name), None)

        if database_enum == None:
            continue

        for value in enum.values:
            if value not in database_enum.values:
                actions.append(EnumAction(type="ADD_ENUM_VALUE", enum=database_enum, value=value))


    # Then look for tables that exist in the model_schema that are not in the database_schema and thus
    # need to be created
    for model_table in model_tables:
        if model_table.name not in [table.name for table in database_tables]:
            actions.append(TableAction(type="CREATE_TABLE", table=model_table))
            # Find indexes that don't exist in the database
            for index in model_table.indexes:
                actions.append(TableAction(type="CREATE_INDEX", table=model_table, index=index))
    
    # Look to see if any tables exist in the database_tables that do not exist in the model_tables
    # And thus need to be dropped
    for db_table in database_tables:
        if db_table.name not in [table.name for table in model_tables]:
            actions.append(TableAction(type="DROP_TABLE", table=db_table))


    for table in model_tables:
        database_table = next(
            (t for t in database_tables if t.name == table.name), None
        )
        if database_table is None:
            continue

        # Find indexes that don't exist in the database
        for index in table.indexes:
            database_index = next((i for i in database_table.indexes if i == index), None)
            if database_index == None:
                actions.append(TableAction(type="CREATE_INDEX", table=table, index=index))

        for db_index in database_table.indexes:
            model_index = next((i for i in table.indexes if i == db_index), None)
            if model_index == None:
                actions.append(TableAction(type="DROP_INDEX", table=database_table, index=db_index))

        # Find columns that exist in the database but do not exist in the model schema anymore
        for db_column in database_table.columns:
            if db_column.name not in [column.name for column in table.columns]:
                actions.append(
                    TableAction(type="DROP_COLUMN", table=db_table, column=db_column)
                )

        # Find columns that exist in the model schema but do not exist in the database schema
        for model_column in table.columns:
            if model_column.name not in [
                column.name for column in database_table.columns
            ]:
                actions.append(
                    TableAction(type="CREATE_COLUMN", table=table, column=model_column)
                )

        # Figure out how a column changed from what is in the DB
        for column in table.columns:
            database_column = next(
                (c for c in database_table.columns if c.name == column.name), None
            )
            if database_column == None:
                continue

            if column.data_type != database_column.data_type:
                actions.append(
                    TableAction(type="CHANGE_DATA_TYPE", table=table, column=column)
                )

            if (column.default or "").lower() != (database_column.default or "").lower():
                actions.append(
                    TableAction(type="COLUMN_DEFAULT", table=table, column=column)
                )

            if column.nullable == True and database_column.nullable == False:
                actions.append(
                    TableAction(type="COLUMN_NULLABLE", table=table, column=column)
                )

            if column.nullable == False and database_column.nullable == True:
                actions.append(
                    TableAction(type="COLUMN_NOT_NULLABLE", table=table, column=column)
                )

            for constraint in column.constraints:
                database_constraint = next((c for c in database_column.constraints if c.__class__ == constraint.__class__), None)

                if database_constraint == None:
                    actions.append(TableAction(type="ADD_CONSTRAINT", table=table, column=column, constraint=constraint))

            for constraint in database_column.constraints:
                model_constraint = next((c for c in column.constraints if c.__class__ == constraint.__class__), None)
                if model_constraint == None:
                    actions.append(TableAction(type="DROP_CONSTRAINT", table=table, column=column, constraint=constraint))

            for constraint in column.constraints:
                database_constraint = next((c for c in database_column.constraints if c.__class__ == constraint.__class__), None)
                if database_constraint == None:
                    continue

                if constraint != database_constraint:
                    actions.append(TableAction(type="DROP_CONSTRAINT", table=table, column=column, constraint=database_constraint))
                    actions.append(TableAction(type="ADD_CONSTRAINT", table=table, column=column, constraint=constraint))

    return actions



from typing import Generic, TypeVar, Dict, Optional

Val = TypeVar('Val')

class DotDict(Generic[Val]):
    _data: Dict[str, Val]
    def __init__(self, **kwargs: Val):
        # Initialize the dictionary with keyword arguments
        self.__dict__["_data"] = kwargs

    def __getitem__(self, key: str) -> Val:
        # Access items like a dictionary
        if key not in self._data:
            raise KeyError(f"Key '{key}' not found in DotDict.")
        return self._data[key]

    def __setitem__(self, key: str, value: Val) -> None:
        # Set items like a dictionary
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        # Delete items like a dictionary
        if key not in self._data:
            raise KeyError(f"Key '{key}' not found in DotDict.")
        del self._data[key]

    def __getattr__(self, key: str) -> Val:
        # Access attributes like dot notation
        try:
            return self._data[key]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __setattr__(self, key: str, value: Val) -> None:
        # Set attributes like dot notation
        if key == "_data":  # Avoid overriding internal data
            super().__setattr__(key, value)
        else:
            self._data[key] = value

    def __delattr__(self, key: str) -> None:
        # Delete attributes like dot notation
        if key in self._data:
            del self._data[key]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __repr__(self) -> str:
        return f"DotDict({self._data!r})"

    def __str__(self) -> str:
        return str(self._data)

    def get(self, key: str, default: Optional[Val] = None) -> Optional[Val]:
        # Provide a dictionary-like 'get' method
        return self._data.get(key, default)

    def keys(self):
        # Return dictionary keys
        return self._data.keys()

    def values(self):
        # Return dictionary values
        return self._data.values()

    def items(self):
        # Return dictionary items
        return self._data.items()



from typing import (
    get_origin,
    Annotated,
    get_args,
    Optional,
    Any,
    Type,
    Generic,
    List,
    Tuple,
    TypeVar,
    Mapping,
)
import re
from asyncpg import Connection
from .query_builder.conditions import Condition, LogicalCondition, AndCondition
from .query_builder.query_builder import QueryBuilder, OrderByDirection
from .dot_dict import DotDict
from .query_builder.model_column import ModelColumn


def convert_named_to_positional(sql: str, params: dict):
    # Find all named parameters
    param_names = re.findall(r":(\w+)", sql)
    # Replace named parameters with positional ones like $1, $2, ...
    transformed_sql = sql
    for i, name in enumerate(param_names):
        transformed_sql = transformed_sql.replace(f":{name}", f"${i+1}")
    # Reorder the parameters according to their position in the query
    transformed_params = [params[name] for name in param_names]
    return transformed_sql, transformed_params


T = TypeVar("T", bound="Model")


class MetaModel(type):
    def __init__(cls: Type["Model"], name, bases, dict):
        super().__init__(name, bases, dict)  # type: ignore

        cls.columns: DotDict[ModelColumn] = DotDict()
        for name in cls.__annotations__.keys():
            cls.columns[name] = ModelColumn(table=cls.__table_name__, name=name)


CreateT = TypeVar("CreateT", bound=Mapping[str, Any])
UpdateT = TypeVar("UpdateT", bound=Mapping[str, Any])


# class Model:
class Model(Generic[CreateT, UpdateT], metaclass=MetaModel):
    __table_name__: str = ""
    __indexes__ = []

    _primary_key: str | None = None
    _required_fields_to_insert: List[str] | None = None
    _updated_at_columns: List[str] | None = None

    @classmethod
    def builder(cls: Type[T]) -> QueryBuilder[T]:
        query_builder = QueryBuilder()
        query_builder.table = cls.__table_name__
        query_builder.return_as(cls)
        return query_builder

    @classmethod
    def create_test[T](cls: T, data: T):
        pass

    @classmethod
    def _get_updated_at_columns(cls):
        if cls._updated_at_columns != None:
            return cls._updated_at_columns

        columns = []
        for column_name, type in cls.__annotations__.items():
            is_annotated = get_origin(type) is Annotated
            if is_annotated == False:
                continue
            _, *annotations = get_args(type)
            if {"updated_at": True} in annotations:
                columns.append(column_name)

        cls._updated_at_columns = columns
        return cls._updated_at_columns

    @classmethod
    def _get_primary_key(cls) -> str:
        if cls._primary_key != None:
            return cls._primary_key

        for column_name, type in cls.__annotations__.items():
            is_annotated = get_origin(type) is Annotated
            if is_annotated == False:
                continue
            _, *annotations = get_args(type)
            if {"primary_key": True} in annotations:
                cls._primary_key = column_name
                break

        return cls._primary_key or ""

    @classmethod
    def _get_required_fields_to_insert(cls: Type[T]):
        if cls._required_fields_to_insert != None:
            return cls._required_fields_to_insert

        field_names = []
        for column_name, type in cls.__annotations__.items():
            is_annotated = get_origin(type) is Annotated

            if is_annotated == False:
                if get_origin(type) != Optional:
                    field_names.append(column_name)
                    continue
            else:
                data_type, *annotations = get_args(type)
                if get_origin(data_type) is Optional or any(
                    "nullable" in annotation
                    or "default" in annotation
                    or "auto_increment" in annotation
                    for annotation in annotations
                ):
                    continue
                else:
                    field_names.append(column_name)
        cls._required_fields_to_insert = field_names
        return cls._required_fields_to_insert

    @classmethod
    async def get(
        cls: Type[T],
        *conditions: LogicalCondition,
        conn: Connection | None = None,
    ) -> T | None:
        result: List[T] = (
            await QueryBuilder()
            .select(cls.__table_name__)
            .where(AndCondition(*list(conditions)))
            .limit(1)
            .return_as(cls)
            .run(conn)
        )
        if len(result) == 0:
            return None
        return result[0]

    @classmethod
    async def query(
        cls: Type[T],
        condition: LogicalCondition | None = None,
        order_by: List[Tuple["ModelColumn", OrderByDirection]] | None = None,
        limit: int | None = None,
        conn: Connection | None = None,
    ):
        query_builder = QueryBuilder().select(cls.__table_name__)

        if condition != None:
            query_builder = query_builder.where(condition)

        if order_by != None:
            query_builder = query_builder.order_by(*order_by)

        if limit != None:
            query_builder = query_builder.limit(limit)

        query_builder = query_builder.return_as(cls)

        result: List[T] = await query_builder.run(conn)
        return result

    @classmethod
    async def create(cls: Type[T], data: CreateT, conn: Connection | None = None) -> T:
        results: List[T] = (
            await QueryBuilder()
            .insert(cls.__table_name__, dict(data))
            .return_as(cls)
            .run(conn)
        )
        return results[0]

    @classmethod
    async def create_many(
        cls: Type[T], data: List[CreateT], conn: Connection | None = None
    ) -> List[T]:
        new_data = [dict(d) for d in data]
        results: List[T] = (
            await QueryBuilder()
            .insert(cls.__table_name__, new_data)
            .return_as(cls)
            .run(conn)
        )
        return results

    @classmethod
    async def update(
        cls: Type[T],
        data: UpdateT,
        condition: LogicalCondition,
        conn: Connection | None = None,
    ):
        result: List[T] = (
            await QueryBuilder()
            .update(cls.__table_name__, dict(data))
            .where(condition)
            .return_as(cls)
            .run(conn)
        )
        return result

    @classmethod
    async def upsert(
        cls: Type[T],
        where: LogicalCondition,
        create: CreateT,
        update: UpdateT,
        conn: Connection | None = None,
    ) -> T:
        existing = (
            await QueryBuilder()
            .select(cls.__table_name__)
            .where(where)
            .return_as(cls)
            .run(conn)
        )
        if len(existing) > 1:
            raise Exception("Where condition for upsert was not unique")

        if len(existing) == 1:
            if len(update.keys()) == 0:
                return existing[0]
            
            result = (
                await QueryBuilder()
                .update(cls.__table_name__, data=dict(update))
                .where(where)
                .return_as(cls)
                .run()
            )
            return result[0]
        else:
            result = (
                await QueryBuilder()
                .insert(cls.__table_name__, data=dict(create))
                .return_as(cls)
                .run()
            )
            return result[0]

    async def update_self(self: T, conn: Connection | None = None) -> T:
        primary_key_column = self.__class__._get_primary_key()
        data = {}
        for key in self.__annotations__.keys():
            if primary_key_column == key:
                continue
            data[key] = getattr(self, key)

        result: List[T] = (
            await QueryBuilder()
            .update(self.__class__.__table_name__, data)
            .where(
                self.columns[primary_key_column] == getattr(self, primary_key_column)
            )
            .return_as(self.__class__)
            .run(conn)
        )
        return result[0]

    @classmethod
    async def delete(
        cls: Type[T], *conditions: LogicalCondition, conn: Connection | None = None
    ):
        await QueryBuilder().delete(cls.__table_name__).where(
            AndCondition(*list(conditions))
        ).run()

    async def delete_self(self, conn: Connection | None = None):
        primary_key_column = self.__class__._get_primary_key()
        primary_key_value = getattr(self, primary_key_column)
        await QueryBuilder().delete(self.__class__.__table_name__).where(
            self.__class__.columns[primary_key_column] == primary_key_value
        ).run(conn)

    # @classmethod
    # async def upsert[
    #     T
    # ](
    #     cls: Type[T],
    #     model: T,
    #     unique_columns: List[str] | None = None,
    #     conn: Connection | None = None,
    # ):
    #     async with _get_connection(conn) as conn:
    #         updated_at_columns = cls._get_updated_at_columns()
    #         for column_name in updated_at_columns:
    #             setattr(
    #                 model, column_name, datetime.datetime.now(datetime.timezone.utc)
    #             )

    #         fields_to_insert = {}
    #         required_fields = cls._get_required_fields_to_insert()
    #         for property in fields(model):
    #             if (
    #                 property.name in required_fields
    #                 or getattr(model, property.name) != None
    #             ):
    #                 fields_to_insert[property.name] = getattr(model, property.name)

    #         if unique_columns == None:
    #             unique_indexes = [
    #                 index.columns
    #                 for index in cls.__indexes__
    #                 if isinstance(index, UniqueIndex)
    #                 and all(key in fields_to_insert for key in index.columns)
    #             ]
    #             unique_columns = next(
    #                 iter(unique_indexes),
    #                 [
    #                     (
    #                         cls._get_primary_key()
    #                         if cls._get_primary_key() in fields_to_insert
    #                         else None
    #                     )
    #                 ],
    #             )
    #             if unique_columns == None:
    #                 raise Exception(
    #                     "No unique columns intersect with the provided data", cls, model
    #                 )

    #         result = await conn.fetch(
    #             f"""
    #             INSERT INTO {cls.__table_name__}
    #             ({",".join([key for key in fields_to_insert.keys()])})
    #             VALUES
    #             ({",".join([f"${i}" for i in range(1, len(fields_to_insert.keys()) + 1)])})
    #             ON CONFLICT ({",".join(unique_columns)})
    #             DO UPDATE SET
    #                 {",\n".join([f"{key} = ${i+1}" for i, key in enumerate(fields_to_insert.keys())])}
    #             RETURNING {", ".join([name for name in cls.__annotations__.keys()])}
    #         """,
    #             *[value for value in fields_to_insert.values()],
    #         )

    #         return cls(**result[0])



import click
from typing import List
from time import time

def write_migration(sql: List[str]):
    ms_date_time = int(time() * 1000)
    name = click.prompt("Migration Name", type=str)
    file_name = f"database/migrations/{ms_date_time}_{name}.py"
    with open(file_name, "w") as file:
        file.write(f"""
# Migration Name: {name}
# Created at: {ms_date_time}
from asyncpg import Connection

async def migrate(conn: Connection):
    {'\n    '.join([f"await conn.execute(\"{statement}\")" for statement in sql])}
                   
""")


from ..utils.get_database_url import get_database_url
from ..utils.run_migrations import run_migrations

async def migrate_prod_command():
    database_url = get_database_url()

    await run_migrations(database_url)


import click
from ..utils.get_database_url import get_database_url
from ..utils.get_models import get_models
from ..utils.model_schema import get_models_schema
from ..utils.database_schema import get_database_schema
from ..utils.schema_differ import get_schema_diff_actions
from ..utils.action import action_to_sql
from ..utils.write_migration import write_migration
from ..utils.run_migrations import run_migrations

async def migrate_dev_command():
    database_url = get_database_url()

    if database_url == None:
        raise Exception("DATABASE_URL is not defined in .env")

    await run_migrations(database_url)

    database_schema = await get_database_schema(database_url)
    model_schema = get_models_schema()

    actions = get_schema_diff_actions(model_schema=model_schema, database_schema=database_schema)

    if len(actions) == 0:
        click.echo("Database is up to date with schema")
        return

    actions_sql = [action_to_sql(action) for action in actions]

    print("Pending migration:")
    for sql in actions_sql:
        print(sql)

    write_migration([sql for sql in actions_sql if sql != None])
    await run_migrations(database_url)


import click
import asyncpg
from urllib.parse import urlparse, urlunparse
from ..utils.get_database_url import get_database_url
from ..utils.run_migrations import run_migrations

async def migrate_reset_command():
    confirmed = click.confirm("Are you sure you want to reset the database?", True)
    if confirmed == False:
        return

    database_url = get_database_url()
    if database_url == None:
        raise Exception("DATABASE_URL is not set in .env")
    
    parsed_url = urlparse(database_url)

    database = parsed_url.path.lstrip("/")

    connection_url = urlunparse(
        (
            parsed_url.scheme,
            parsed_url.netloc,
            "/",
            "",
            "",
            "",
        )
    )

    conn = await asyncpg.connect(connection_url)
    await conn.execute(f'DROP DATABASE IF EXISTS "{database}"')
    await conn.execute(f'CREATE DATABASE "{database}"')
    await conn.close()

    await run_migrations(database_url)

    click.echo("Database reset")



import os
import importlib

def dynamic_import(file_path):
    # Get the absolute path of the file
    file_path = os.path.abspath(file_path)

    # Create a module name from the file name (without extension)
    module_name = os.path.splitext(os.path.basename(file_path))[0]

    # Load the module dynamically
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


import secrets
import string

# Define Nano ID–like alphabet
alphabet = string.ascii_letters

def nanoid():
    # Generate a 21-character ID
    return ''.join(secrets.choice(alphabet) for _ in range(10))


import asyncio
import click
from .commands.init import init_command
from .commands.migrate_dev import migrate_dev_command
from .commands.migrate_prod import migrate_prod_command
from .commands.migrate_create import migrate_create_command
from .commands.migrate_reset import migrate_reset_command

@click.group()
def main():
    pass

@main.command()
def init():
    asyncio.run(init_command())

@main.group()
def migrate():
    pass

@migrate.command()
def dev():
    asyncio.run(migrate_dev_command())

@migrate.command()
def prod():
    asyncio.run(migrate_prod_command())

@migrate.command()
def create():
    asyncio.run(migrate_create_command())

@migrate.command()
def reset():
    asyncio.run(migrate_reset_command())

if __name__ == "__main__":
    main()


from typing import (
    Type,
    List,
    Any,
    get_origin,
    Union,
    get_args,
    Annotated,
    Dict,
)
import os
import sys
import importlib
from enum import StrEnum
from dataclasses import dataclass
from .to_snake_case import to_snake_case
from .schema import Table, Column, ForeignKeyConstraint, UniqueConstraint, Index, Enum
from ...model import Model

DATA_TYPE_LOOKUP = {
    "str": "text",
    "int": "int4",
    "bool": "bool",
    "float": "double precision",
    "datetime": "timestamptz(3)",
    "dict": "jsonb",
    "list": "jsonb",
}

# Add the FastAPI project's root directory to sys.path
project_root = os.getcwd()
if project_root not in sys.path:
    sys.path.insert(0, project_root)


def get_relavent_exports():
    model_files = os.listdir("database/models")

    exports = set()
    for model_file in model_files:
        if model_file.endswith(".py") == False:
            continue

        module = importlib.import_module(
            f"database.models.{model_file.replace(".py", "")}"
        )
        for attribute_name in dir(module):
            export = getattr(module, attribute_name)
            if (
                isinstance(export, type)
                and export != Model
                and (issubclass(export, Model) or issubclass(export, StrEnum))
            ):
                exports.add(export)

    return exports


def enum_name(enum: Type[StrEnum]):
    return to_snake_case(enum.__name__)


def is_optional(type_hint):
    origin = get_origin(type_hint)  # Get the origin of the type (e.g., Union)
    return origin is Union and None.__class__ in get_args(type_hint)


class ColumnDataTypeNotCompatible(Exception):
    def __init__(self, name: str, model_name: str):
        self.name = name

        super().__init__(f"The type for {name} on model {model_name} is incompatible.")


def get_column_type(python_type):
    origin = get_origin(python_type)

    if origin is Union:
        python_type = [type for type in get_args(python_type) if type is not None][0]

    if issubclass(python_type, StrEnum):
        return enum_name(python_type)

    type = DATA_TYPE_LOOKUP[python_type.__name__]

    if type == None:
        raise
    return type


def is_enum_type(python_type):
    origin = get_origin(python_type)
    if origin is Union:
        python_type = next(
            (type for type in get_args(python_type) if type is not None), None
        )

    if python_type is None:
        raise Exception(f"Python type could not be determined")

    if issubclass(python_type, StrEnum):
        return python_type

    return None


def get_table_name(model: Type[Model]):
    return model.__table_name__ or to_snake_case(model.__name__)


def get_model_schema(model: Type[Model]):
    columns = []
    for column_name, type in model.__annotations__.items():
        is_annotated = get_origin(type) is Annotated
        data_type = type
        meta_data: Dict[str, Any] = {}
        if is_annotated:
            data_type, *annotations = get_args(type)
            meta_data["data_type"] = get_column_type(data_type)
            for annotation in annotations:
                meta_data.update(annotation)
            meta_data["nullable"] = (
                is_optional(data_type)
                and "default" not in meta_data
                and "primary_key" not in meta_data
                and "auto_increment" not in meta_data
            )
        else:
            meta_data["data_type"] = get_column_type(data_type)
            meta_data["nullable"] = is_optional(data_type)

        constraints = []
        foreign_key = meta_data.get("foreign_key")
        unique: bool = meta_data.get("unique", False)
        if foreign_key != None:
            constraints.append(ForeignKeyConstraint(references=foreign_key.get("key"), on_delete=foreign_key.get("on_delete")))
        if unique:
            constraints.append(UniqueConstraint())

        columns.append(
            Column(
                name=column_name,
                data_type=meta_data.get("data_type", ""),
                nullable=meta_data.get("nullable", False),
                primary_key=meta_data.get("primary_key", False),
                auto_increment=meta_data.get("auto_increment", False),
                default=meta_data.get("default", None),
                constraints=constraints,
            )
        )

    indexes = []
    for index in model.__indexes__:
        indexes.append(
            Index(
                name=f"idx_{get_table_name(model)}_{index.type}_{"_".join([column for column in index.columns])}",
                type=index.type,
                columns=index.columns,
            )
        )

    table = Table(name=get_table_name(model), columns=columns, indexes=indexes)
    return table


def get_models_schema():
    exports = get_relavent_exports()

    tables = []
    enums = []
    for export in exports:
        if issubclass(export, Model):
            tables.append(get_model_schema(export))
        elif issubclass(export, StrEnum) and export != StrEnum:
            enums.append(
                Enum(name=enum_name(export), values=[member.value for member in export])
            )

    return tables, enums


from typing import Optional
from dataclasses import dataclass
from .schema import (
    Table,
    Column,
    Constraint,
    ForeignKeyConstraint,
    UniqueConstraint,
    Index,
    Enum,
)

@dataclass
class EnumAction:
    type: str

    enum: Enum
    value: Optional[str] = None

@dataclass
class TableAction:
    type: str

    table: Table
    index: Optional[Index] = None
    column: Optional[Column] = None
    constraint: Optional[Constraint] = None


def create_column_sql(column: Column):
    foreign_key_constraint = next(
        (
            constraint
            for constraint in column.constraints
            if isinstance(constraint, ForeignKeyConstraint)
        ),
        None,
    )
    is_unique = next(
        (
            constraint
            for constraint in column.constraints
            if isinstance(constraint, UniqueConstraint)
        ),
        None,
    )
    return (
        f"{column.name}"
        + (f" {column.data_type}" if column.auto_increment == False else " SERIAL")
        + (" NOT NULL" if not column.nullable and not column.primary_key else "")
        + (" PRIMARY KEY" if column.primary_key else "")
        + (f" DEFAULT {column.default}" if column.default else "")
        + (f" UNIQUE" if is_unique else "")
        (
            f" REFERENCES {foreign_key_constraint.references} ON DELETE {foreign_key_constraint.on_delete}"
            if foreign_key_constraint and foreign_key_constraint.on_delete
            else ""
        )
        + (
            f" REFERENCES {foreign_key_constraint.references}"
            if foreign_key_constraint
            else ""
        )
    )


def action_to_sql(action: TableAction | EnumAction) -> str | None:
    if isinstance(action, EnumAction):
        match action.type:
            case "CREATE_ENUM":
                values_sql = ", ".join([f"'{value}'" for value in action.enum.values])
                return f"CREATE TYPE {action.enum.name} AS ENUM ({values_sql})"
            case "DROP_ENUM":
                return f"DROP TYPE {action.enum.name}"
            case "ADD_ENUM_VALUE":
                return f"ALTER TYPE {action.enum.name} ADD VALUE '{action.value}'"


    if isinstance(action, TableAction):
        match action.type:
            case "DROP_TABLE":
                return f"DROP TABLE {action.table.name}"
            case "CREATE_TABLE":
                columns_sql = ", ".join(
                    [create_column_sql(column) for column in action.table.columns]
                )
                return f"CREATE TABLE {action.table.name} ({columns_sql})"
            case "CREATE_INDEX":
                if action.index == None:
                    raise Exception("Index was not provided on the action")
                return f"CREATE{" UNIQUE" if action.index.type == 'unique' else ""} INDEX {action.index.name} ON {action.table.name} ({",".join(action.index.columns)})"
            case "DROP_INDEX":
                if action.index == None:
                    raise Exception("Index was not provided on the action")
                
                return f"DROP INDEX {action.index.name}"
            case "DROP_COLUMN":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                return f"ALTER TABLE {action.table.name} DROP COLUMN {action.column.name}"
            case "CREATE_COLUMN":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                return f"ALTER TABLE {action.table.name} ADD COLUMN {create_column_sql(action.column)}"
            case "CHANGE_DATA_TYPE":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                old_column_sql = f"ALTER TABLE {action.table.name} ALTER COLUMN {action.column.name} TYPE {action.column.data_type}"
                if action.column.data_type == "uuid":
                    old_column_sql += f" USING {action.column.name}::uuid"
                return old_column_sql
            case "COLUMN_NULLABLE":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                return f"ALTER TABLE {action.table.name} ALTER COLUMN {action.column.name} DROP NOT NULL"

            case "COLUMN_NOT_NULLABLE":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                return f"ALTER TABLE {action.table.name} ALTER COLUMN {action.column.name} SET NOT NULL"
            case "COLUMN_DEFAULT":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                if action.column.default == None:
                    return f"ALTER TABLE {action.table.name} ALTER COLUMN {action.column.name} DROP DEFAULT"
                else:
                    return f"ALTER TABLE {action.table.name} ALTER COLUMN {action.column.name} SET DEFAULT {action.column.default}"
            case "ADD_CONSTRAINT":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                if isinstance(action.constraint, ForeignKeyConstraint):
                    sql = f"ALTER TABLE {action.table.name} ADD CONSTRAINT fk_{action.column.name} FOREIGN KEY ({action.column.name}) REFERENCES {action.constraint.references}"
                    if action.constraint.on_delete != None:
                        sql += f" ON DELETE {action.constraint.on_delete}"
                    return sql
                elif isinstance(action.constraint, UniqueConstraint):
                    return f"ALTER TABLE {action.table.name} ADD CONSTRAINT uq_{action.column.name} UNIQUE ({action.column.name})"
            case "DROP_CONSTRAINT":
                if action.column == None:
                    raise Exception("Column was not provided on the action")
                if isinstance(action.constraint, ForeignKeyConstraint):
                    return f"ALTER TABLE {action.table.name} DROP CONSTRAINT fk_{action.column.name}"
                elif isinstance(action.constraint, UniqueConstraint):
                    return f"ALTER TABLE {action.table.name} DROP CONSTRAINT uq_{action.column.name}"
            case _:
                raise ValueError(f"Unknown Action type: {action.type}")



from typing import Any, Union, List
from dataclasses import dataclass
from .model_column import ModelColumn
from ..nanoid import nanoid

LogicalCondition = Union["Condition", "AndCondition", "OrCondition"]

@dataclass
class Condition:
    column: ModelColumn
    condition: str
    value: Any

    def to_sql(self):
        if isinstance(self.value, ModelColumn):
            return (
                f"{self.column.table}.{self.column.name} {self.condition} {self.value.table}.{self.value.name}",
                {},
            )

        parameter_name = f"condition_{self.column.table}_{self.column.name}_{nanoid()}"

        if self.condition == "in":
            return f"{self.column.table}.{self.column.name} = ANY(:{parameter_name})", {
                parameter_name: self.value
            }

        return (
            f"{self.column.table}.{self.column.name} {self.condition} :{parameter_name}",
            {parameter_name: self.value},
        )

@dataclass
class OrCondition:
    conditions: List[LogicalCondition]

    def __init__(self, *conditions: LogicalCondition):
        self.conditions = list(conditions)

    def to_sql(self):
        parameters = {}
        sql_statements = []
        for condition in self.conditions:
            condition_sql, condition_parameters = condition.to_sql()
            sql_statements.append(condition_sql)
            parameters.update(condition_parameters)
        return f"({" OR ".join(sql_statements)})", parameters


@dataclass
class AndCondition:
    conditions: List[LogicalCondition]

    def __init__(self, *conditions: LogicalCondition):
        self.conditions = list(conditions)

    def to_sql(self):
        parameters = {}
        sql_statements = []
        for condition in self.conditions:
            condition_sql, condition_parameters = condition.to_sql()
            sql_statements.append(condition_sql)
            parameters.update(condition_parameters)
        return f"({" AND ".join(sql_statements)})", parameters




from .get_models import get_models
from .model_schema import get_model_schema
from .schema_differ import get_schema_diff_actions
from .action import action_to_sql
from .run_migrations import run_migrations


import re

def to_snake_case(class_name: str) -> str:
    # Use a regular expression to add underscores between lowercase and uppercase letters
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', class_name).lower()



from typing import List, Optional, Tuple, Dict, TypeVar, Type, Generic
from contextlib import asynccontextmanager
from dataclasses import dataclass
from enum import StrEnum, auto
from asyncpg import Connection
import re
from .. import get_connection
from .model_column import ModelColumn
from .conditions import LogicalCondition, Condition
from ..nanoid import nanoid

def convert_named_to_positional(sql: str, params: dict):
    # Find all named parameters
    param_names = re.findall(r":(\w+)", sql)
    # Replace named parameters with positional ones like $1, $2, ...
    transformed_sql = sql
    for i, name in enumerate(param_names):
        transformed_sql = transformed_sql.replace(f":{name}", f"${i+1}")
    # Reorder the parameters according to their position in the query
    transformed_params = [params[name] for name in param_names]
    return transformed_sql, transformed_params

class OrderByDirection(StrEnum):
    ASC = auto()
    DESC = auto()


class QueryType(StrEnum):
    SELECT = auto()
    INSERT = auto()
    UPDATE = auto()
    DELETE = auto()

@dataclass
class Join:
    type: str
    table: str
    condition: LogicalCondition

    def to_sql(self):
        condition_sql, parameters = self.condition.to_sql()
        return f"{self.type} {self.table} ON {condition_sql}", parameters

T = TypeVar('T')

class QueryBuilder(Generic[T]):
    query_type: QueryType | None
    table: str | None
    is_distinct: bool
    joins: List[Join]
    conditions: List[LogicalCondition]
    limit_value: Optional[int]
    order_by_conditions: List[Tuple[ModelColumn, OrderByDirection] | str]
    return_columns: List[str]
    return_as_cls: Type[T] | None
    data: List[Dict] | Dict | None

    def __init__(self):
        self.return_model = None
        self.query_type = None
        self.table = None
        self.is_distinct = False
        self.joins = []
        self.conditions = []
        self.limit_value = None
        self.order_by_conditions = []
        self.return_columns = []
        self.return_as_cls = None
        self.data = None

    def select(self, table: str | None = None, columns: List[str] | None = None):
        self.query_type = QueryType.SELECT
        self.table = table or self.table
        self.return_columns = columns or self.return_columns or []

        return self

    def insert(self, table: str, data: List[Dict] | Dict):
        self.query_type = QueryType.INSERT
        self.table = table
        self.data = data
        return self

    def update(self, table: str, data: List[Dict] | Dict):
        self.query_type = QueryType.UPDATE
        self.table = table
        self.data = data
        return self

    def delete(self, table: str):
        self.query_type = QueryType.DELETE
        self.table = table
        return self
    
    def distinct(self):
        self.is_distinct = True
        return self

    def where(self, condition: LogicalCondition):
        self.conditions.append(condition)
        return self

    def join(
        self, table: str, condition: LogicalCondition, type: str = "JOIN"
    ):
        self.joins.append(Join(type, table, condition))
        return self

    def left_join(self, table: str, condition: LogicalCondition):
        self.join(table, condition, "LEFT JOIN")
        return self

    def right_join(self, table: str, condition: LogicalCondition):
        self.join(table, condition, "RIGHT JOIN")
        return self

    def inner_join(self, table: str, condition: LogicalCondition):
        self.join(table, condition, "INNER JOIN")
        return self

    def outer_join(self, table: str, condition: LogicalCondition):
        self.join(table, condition, "OUTER JOIN")
        return self

    def limit(self, limit: int):
        self.limit_value = limit
        return self

    def order_by(self, *columns: Tuple["ModelColumn", OrderByDirection] | str):
        self.order_by_conditions += columns
        return self
    
    def return_as(self, model_cls: Type[T]):
        self.return_as_cls = model_cls
        self.return_columns = list(model_cls.__annotations__.keys())
        return self

    def select_sql(self):
        if len(self.return_columns) == 0:
            raise Exception("No columns selected to return")

        parameters = {}
        query = f"SELECT "
        query += "DISTINCT " if self.is_distinct else ""
        query += ", ".join([f"{self.table}.{column}" for column in self.return_columns])
        query += f" FROM {self.table}"

        if len(self.joins) > 0:
            query += " "
            join_sql = []
            for join in self.joins:
                sql, join_params = join.to_sql()
                join_sql.append(sql)
                parameters.update(join_params)
            query += " ".join(join_sql)

        if len(self.conditions) > 0:
            query += " WHERE "
            where_sql = []
            for where_condition in self.conditions:
                sql, params = where_condition.to_sql()
                where_sql.append(sql)
                parameters.update(params)
            query += " AND ".join(where_sql)
        if self.limit_value != None:
            limit_param_name = f"limit_{nanoid()}"
            query += f" LIMIT :{limit_param_name}"
            parameters[limit_param_name] = self.limit_value
        if len(self.order_by_conditions) > 0:
            query += " ORDER BY "
            order_by = []
            for o in self.order_by_conditions:
                if isinstance(o, str):
                    order_by.append(o)
                else:
                    column, direction = o
                    order_by.append(f"{column.table}.{column.name} {direction}")
            query += ", ".join(order_by)

        return convert_named_to_positional(query, parameters)

    def insert_sql(self):
        if self.table == None:
            raise Exception("No table was selected to insert into")
        if self.data == None:
            raise Exception("No data provided to insert")
        
        query: str = "INSERT INTO "
        query += self.table
        
        columns = self.data[0].keys() if isinstance(self.data, list) else self.data.keys()
        query += f" ({", ".join(columns)})"
        
        parameters = {}
        rows_sql = []
        if isinstance(self.data, list):
            for row in self.data:
                row_sql = []
                for i, column in enumerate(columns):
                    parameter_name = f"row_{i}_{column}_{nanoid()}"
                    parameters[parameter_name] = row[column]
                    row_sql.append(f":{parameter_name}")
                rows_sql.append(f"({", ".join(row_sql)})")
        else:
            row_sql = []
            for column in columns:
                parameter_name = f"row_{column}_{nanoid()}"
                parameters[parameter_name] = self.data[column]
                row_sql.append(f":{parameter_name}")
            rows_sql.append(f"({", ".join(row_sql)})")

        query += " VALUES "
        query += ", ".join(rows_sql)


        if self.return_as_cls != None:
            query += f" RETURNING {", ".join([name for name in self.return_as_cls.__annotations__.keys()])}"

        return convert_named_to_positional(query, parameters)

    def update_sql(self):
        if self.data is None:
            raise Exception("Update data can not be None")
        if isinstance(self.data, list):
            raise Exception("Update statement can't have a list of data")

        query = f"UPDATE {self.table}"
        query += " SET "
        
        parameters = {}
        update_sql = []
        for key, value in self.data.items():
            parameter_name = f"column_{self.table}_{key}_{nanoid()}"
            update_sql.append(f"{key} = :{parameter_name}")
            parameters[parameter_name] = value
        query += ", ".join(update_sql)

        query += " WHERE "

        

        conditions_sql = []
        for condition in self.conditions:
            condition_sql, condition_parameters = condition.to_sql()
            parameters.update(condition_parameters)
            conditions_sql.append(condition_sql)

        query += " AND ".join(conditions_sql)

        if self.return_as_cls != None:
            query += f" RETURNING {", ".join([name for name in self.return_as_cls.__annotations__.keys()])}"
        
        return convert_named_to_positional(query, parameters)

    def delete_sql(self):
        query = f"DELETE FROM {self.table}"
        query += " WHERE "

        parameters = {}
        conditions_sql = []
        for condition in self.conditions:
            sql, params = condition.to_sql()
            conditions_sql.append(sql)
            parameters.update(params)

        query += " AND ".join(conditions_sql)

        return convert_named_to_positional(query, parameters)

    def sql(self):
        match self.query_type:
            case QueryType.SELECT:
                return self.select_sql()
            case QueryType.INSERT:
                return self.insert_sql()
            case QueryType.UPDATE:
                return self.update_sql()
            case QueryType.DELETE:
                return self.delete_sql()
            case _:
                raise Exception(f"Query type {self.query_type} is not implemented")

    async def run(self, conn: Connection | None = None) -> List[T]:
        async with get_connection(conn) as conn:
            sql, params = self.sql()
            results = await conn.fetch(sql, *params)
        if self.return_as_cls == None:
            return results
        else:
            return [self.return_as_cls(**result) for result in results]




def int2():
    return {"data_type": "int2"}  # Small integer

def int4():
    return {"data_type": "int4"}  # Integer

def int8():
    return {"data_type": "int8"}  # Big integer

def float4():
    return {"data_type": "float4"}  # Single precision floating-point

def float8():
    return {"data_type": "float8"}  # Double precision floating-point

def numeric(precision: int | None = None, scale: int | None =None):
    # Numeric with optional precision and scale
    data_type = "numeric"
    if precision is not None and scale is not None:
        data_type += f"({precision}, {scale})"
    return {"data_type": data_type}

def serial():
    return {"data_type": "serial"}  # Auto-incrementing integer

def bigserial():
    return {"data_type": "bigserial"}  # Auto-incrementing big integer

def varchar(length:int):
    return {"data_type": f"varchar({length})"}  # Variable-length with limit

def char(length:int):
    return {"data_type": f"char({length})"}  # Fixed-length

def text():
    return {"data_type": "text"}  # Variable-length, unlimited size

def uuid():
    return {"data_type": "uuid"}

def date():
    return {"data_type": "date"}  # Date only

def timestamp(with_timezone=False, precision:int=3):
    # Timestamp with optional timezone
    return {"data_type": f"timestamptz({precision})" if with_timezone else "timestamp"}

def time(with_timezone=False):
    # Time with optional timezone
    return {"data_type": "timetz" if with_timezone else "time"}

def interval():
    return {"data_type": "interval"}  # Time interval

def data_type(data_type: str):
    return {"data_type": data_type}


from typing import List, Dict
import asyncpg
from .schema import Table, Column, Index, ForeignKeyConstraint, UniqueConstraint, Enum


async def get_database_schema(database_url):
    conn = await asyncpg.connect(database_url)

    tables: List[Table] = []
    enums: List[Enum] = []

    tables_result = await conn.fetch(
        """
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = 'public' AND table_name != '_actual_orm_migrations'
    """
    )

    for table_result in tables_result:
        table_name = table_result["table_name"]

        columns_result = await conn.fetch(
            """
SELECT 
    c.column_name,
    c.udt_name AS data_type,
    c.datetime_precision,
    c.character_maximum_length AS char_length,
    c.is_nullable,
    c.column_default AS default_value,
    COALESCE(pk.constraint_name IS NOT NULL, false) AS is_primary_key,
    COALESCE(s.relname IS NOT NULL AND s.relkind = 'S', false) AS is_auto_increment,
    rcu.table_name as foreign_key_table,
    rcu.column_name as foreign_key_column,
    COALESCE(uc.constraint_name IS NOT NULL, false) AS is_unique_constraint
FROM 
    information_schema.columns c
LEFT JOIN 
    (SELECT kcu.column_name, tc.constraint_name
     FROM 
         information_schema.key_column_usage kcu
     JOIN 
         information_schema.table_constraints tc 
         ON kcu.constraint_name = tc.constraint_name 
         AND kcu.table_schema = tc.table_schema 
     WHERE 
         tc.constraint_type = 'PRIMARY KEY'
    ) pk 
    ON c.column_name = pk.column_name 
    AND c.table_name = $1
LEFT JOIN 
    pg_class s 
    ON c.column_default LIKE '%' || s.relname || '%' 
    AND s.relkind = 'S'
LEFT JOIN 
    information_schema.key_column_usage kcu
    ON c.column_name = kcu.column_name 
    AND c.table_name = kcu.table_name
LEFT JOIN 
    information_schema.referential_constraints rc
    ON kcu.constraint_name = rc.constraint_name
LEFT JOIN 
    information_schema.constraint_column_usage rcu
    ON rc.unique_constraint_name = rcu.constraint_name
    AND kcu.table_schema = rcu.table_schema
LEFT JOIN 
    information_schema.key_column_usage uc_kcu
    ON c.column_name = uc_kcu.column_name
    AND c.table_name = uc_kcu.table_name
LEFT JOIN 
    information_schema.table_constraints uc
    ON uc_kcu.constraint_name = uc.constraint_name
    AND uc.constraint_type = 'UNIQUE'
WHERE 
    c.table_name = $1
AND 
    c.table_schema = 'public';
        """,
            table_name,
        )

        indexes_result = await conn.fetch(r"""
            SELECT
                ix.relname as index_name,
                indisunique as is_unique,
                replace(regexp_replace(regexp_replace(regexp_replace(pg_get_indexdef(indexrelid), ' WHERE .+|INCLUDE .+', ''), ' WITH .+', ''), '.*\((.*)\)', '\1'), ' ', '') AS column_name
            FROM pg_index i 
            JOIN pg_class t ON t.oid = i.indrelid
            JOIN pg_class ix ON ix.oid = i.indexrelid
            JOIN pg_namespace n ON t.relnamespace = n.oid
            WHERE t.relname = $1
                AND n.nspname = 'public';
        """, table_name)

        columns = []
        for column_result in columns_result:
            data_type = column_result['data_type']
            if data_type == 'timestamptz':
                data_type = f'timestamptz({column_result.get('datetime_precision', 3)})'
            if data_type == 'varchar':
                data_type = f'varchar({column_result.get('char_length', 1)})'
            if column_result['is_auto_increment']:
                data_type = 'serial'

            constraints = []
            if column_result['foreign_key_table'] is not None and column_result['foreign_key_column'] is not None:
                constraints.append(ForeignKeyConstraint(references=f"{column_result["foreign_key_table"]}({column_result["foreign_key_column"]})"))

            if column_result['is_unique_constraint']:
                constraints.append(UniqueConstraint())

            column = Column(
                name=column_result["column_name"],
                data_type=data_type,
                nullable=column_result["is_nullable"] == "YES",
                primary_key=column_result["is_primary_key"],
                auto_increment=column_result["is_auto_increment"],
                default=(
                    column_result["default_value"]
                    if column_result["is_auto_increment"] == False
                    else None
                ),
                constraints=constraints
            )

            columns.append(column)
        
        indexes = []
        for index in indexes_result:
            if index.get('index_name', "").startswith("idx_") == False:
                continue
            indexes.append(
                Index(
                    name=index.get('index_name'),
                    type="unique" if index.get('is_unique') else "index",
                    columns=index.get("column_name", "").split(',')
                )
            )

        tables.append(Table(name=table_name, columns=columns, indexes=indexes))

    enums_result = await conn.fetch("""
        SELECT
            n.nspname AS schema_name,
            t.typname AS enum_name,
            e.enumlabel AS enum_value
        FROM
            pg_type t
        JOIN
            pg_enum e ON t.oid = e.enumtypid
        JOIN
            pg_namespace n ON t.typnamespace = n.oid
        ORDER BY
            schema_name, enum_name, e.enumsortorder;                             
    """)
    enums_lookup: Dict[str, Enum] = {}
    for result in enums_result:
        enum = enums_lookup.get(result['enum_name'], Enum(name=result['enum_name'], values=[]))
        enum.values.append(result['enum_value'])
        enums_lookup[result['enum_name']] = enum
    enums = [e for e in enums_lookup.values()]

    await conn.close()
    return tables, enums



import os
import importlib
import asyncpg


async def run_migrations(database_url: str):
    migration_file_names = os.listdir("database/migrations")

    conn = await asyncpg.connect(database_url)
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS _actual_orm_migrations (
                name TEXT,
                ran_at TIMESTAMPTZ(3) DEFAULT NOW()
            );"""
    )
    already_ran = await conn.fetch("""
        SELECT name
        FROM _actual_orm_migrations    
    """)

    for migration_file_name in sorted(migration_file_names):
        if migration_file_name.endswith(".py") == False:
            continue

        migration_name = migration_file_name.replace(".py", "")
        module = importlib.import_module(f"database.migrations.{migration_name}")

        if migration_name in [record["name"] for record in already_ran]:
            continue
        
        # TODO: Keep track of all migrations in the database so we know which ones
        # have been run and which ones have not

        migrate = getattr(module, "migrate")
        async with conn.transaction():
            await migrate(conn)
            await conn.execute("""
                INSERT INTO _actual_orm_migrations
                (name)
                VALUES
                ($1)
            """, migration_name)



from typing import AsyncGenerator
import asyncio
import asyncpg
from contextlib import asynccontextmanager
from weakref import WeakKeyDictionary

# Use WeakKeyDictionary to automatically clean up closed event loops
connection_url: str | None = None
connection_pools = WeakKeyDictionary()

def configure(database_url: str):
    global connection_url
    connection_url = database_url

async def get_or_create_pool() -> asyncpg.Pool:
    global connection_url, connection_pools
    loop = asyncio.get_running_loop()

    # Retrieve or create the pool for the current event loop
    if loop not in connection_pools:
        if not connection_url:
            raise ValueError("Database URL is not configured. Call 'configure()' with a valid URL.")
        
        # Initialize and store a new pool for this event loop
        connection_pools[loop] = await asyncpg.create_pool(connection_url)
    
    return connection_pools[loop]

@asynccontextmanager
async def get_connection(conn: asyncpg.Connection | None = None) -> AsyncGenerator[asyncpg.Connection, None]:
    pool = await get_or_create_pool()
    if conn is None:
        async with pool.acquire() as connection:
            yield connection
    else:
        yield conn

@asynccontextmanager
async def start_transaction(conn: asyncpg.Connection | None = None):
    async with get_connection(conn) as conn:
        async with conn.transaction():
            yield conn

async def close():
    # Close all connection pools and clear the WeakKeyDictionary
    pool = await get_or_create_pool()
    await pool.close()




from typing import List
from dataclasses import dataclass

@dataclass
class Index():
    type = 'index'
    columns: List[str]

@dataclass
class UniqueIndex():
    type = 'unique'
    columns: List[str]


from typing import Callable, TypedDict, get_origin, get_args, Annotated, Dict
from enum import StrEnum, auto
from dataclasses import field, fields
# from .func import now, current_timestamp
from actual_orm.model import Model
from . import data_types

def primary_key():
    return {"primary_key": True}

def auto_increment():
    return {
        "auto_increment": True,
        **data_types.serial(),
    }

def now():
    return "NOW()"

def random():
    return "RANDOM()"

def default(value: str):
    return {
        "default": value
    }

def updated_at():
    return {
        "updated_at": True
    }

def cascade():
    return { "on_delete": "CASCADE" }

def foreign_key(references: type[Model], on_delete: Dict | None = None):
    primary_key = next((column_name for column_name, type in references.__annotations__.items() if get_origin(type) is Annotated and {"primary_key": True} in get_args(type)), None)
    if primary_key is None:
        raise Exception(f"No primary key found on table {references.__class__.__name__}")
    return {
        "foreign_key": {
            "key": f"{references.__table_name__}({primary_key})",
            **(on_delete or {})
        }
    }

def unique():
    return {"unique": True}


import os
import click

def find_requirements_txt(start_dir):
    """Walks up the directory tree starting from start_dir to find requirements.txt."""
    current_dir = start_dir
    while True:
        if "requirements.txt" in os.listdir(current_dir):
            return current_dir
        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:  # This happens when we reach the root directory
            raise FileNotFoundError("No requirements.txt found")
        current_dir = parent_dir

def write_env_file(directory, database_url):
    """Writes the .env file with DATABASE_URL in the specified directory."""
    env_path = os.path.join(directory, ".env")
    with open(env_path, "a" if os.path.isfile(env_path) else "w") as file:
        file.write(f"\nDATABASE_URL={database_url}\n")

async def init_command():
    user = click.prompt("Database user", type=str, default="postgres")
    password = click.prompt("Database password", type=str, default="password")
    host = click.prompt("Database host", type=str, default="localhost")
    port = click.prompt("Database port", type=str, default="5432")
    database = click.prompt("Database name", type=str)

    if os.path.isdir("database/migrations") == False:
        os.makedirs("database/migrations")
    if os.path.isdir("database/models") == False:
        os.makedirs("database/models")

    database_url = f"postgres://{user}:{password}@{host}:{port}/{database}"
    write_env_file(os.getcwd(), database_url)
    